import os,sys
from nose.tools import assert_equal, with_setup

sys.path.append(
    os.path.dirname(os.path.abspath(os.path.join(__file__, os.path.pardir))))

import savfile as savfile
from project import Project

SAV_IN = os.path.join(os.path.dirname(__file__), "test_data", "lsdj.sav")
SAV_OUT = os.path.join(os.path.dirname(__file__), "lsdj.sav.out")

def setup():
   if os.path.exists(SAV_OUT):
       os.unlink(SAV_OUT)

def teardown():
   if os.path.exists(SAV_OUT):
       os.unlink(SAV_OUT)

@with_setup(setup, teardown)
def test_project_save_load():
    print "Loading..."
    sav = savfile.SAVFile(SAV_IN)

    print "Saving..."
    sav.save(SAV_OUT)

    new_sav = savfile.SAVFile(SAV_OUT)

    assert_equal(sav, new_sav)



from utils import add_song_data_property, assert_index_sane, ObjectLookupDict

class Phrase(object):
    """A phrase is a sequence of notes for a single channel.
    """
    def __init__(self, song, index):
        self._song = song
        self._index = index
        self._instruments = ObjectLookupDict(
            self._song.song_data.phrase_instruments[index],
            self._song.instruments)

    @property
    def song(self):
        """a reference to the phrase's parent song"""
        return self._song

    @property
    def index(self):
        """the phrase's index within its parent song's phrase table"""
        return self._index

    @property
    def instruments(self):
        """a list of Instruments, None where no instrument is defined"""
        return self._instruments

for property_name, doc in [
        ("notes", "a list of the phrase's notes, one byte per note"),
        ("fx", "a list of the phrase's effects, one byte per effect"),
        ("fx_val", "a list of the phrase's effect parameters, "
         "one byte per effect")]:
    add_song_data_property(Phrase, property_name, ("phrase_" + property_name,),
                           use_index = True, doc=doc)



import bread_spec
import bread
import os
import sys
from struct import unpack
import utils
from StringIO import StringIO
from project import Project
import blockutils
from blockutils import BlockReader, BlockWriter, BlockFactory
import filepack
import collections
import bitstring

# By default, SAV file loading doesn't trigger any callback action
def _noop_callback(message, step, total_steps, continuing):
    pass

class SAVFile(object):
    # Start offset of SAV file contents
    START_OFFSET = 0x8000

    HEADER_EMPTY_SECTION_1 = (0x8120, 0x813d)

    # Offset of SRAM initialization check
    SRAM_INIT_CHECK_OFFSET = 0x813e
    SRAM_INIT_CHECK_LENGTH = 2

    # Offset where active file number appears
    ACTIVE_FILE_NUMBER_OFFSET = 0x8140

    # Start of block allocation table
    BAT_START_OFFSET = 0x8141

    # End of block allocation table
    BAT_END_OFFSET = 0x81ff

    # Start index for data blocks
    BLOCKS_START_OFFSET = 0x8000

    # The maximum number of files that the .sav can support
    NUM_FILES = 0x20

    # Max length in bytes of filename
    FILENAME_LENGTH = 8

    # Length in bytes of file version number
    FILE_VERSION_LENGTH = 1

    # Length in bytes of file number
    FILE_NUMBER_LENGTH = 1

    #Constants
    EMPTY_BLOCK = 0xff

    def __init__(self, filename, callback=_noop_callback):
        """Constructor.

        :param filename: the file to open
        :type name: str
        :param callback: a progress callback function
        :type name: function
        """
        with open(filename, 'rb') as fp:
            self._load(fp, callback)

    def _load(self, fp, callback):
        # 32 possible projects + read preamble + decompress blocks + "all done"
        total_steps = 35
        current_step = 0

        self.projects = {}

        callback("Reading preamble", current_step, total_steps, True)

        self.preamble = fp.read(self.START_OFFSET)

        header_block_data = fp.read(blockutils.BLOCK_SIZE)

        self.header_block = bread.parse(
            header_block_data, bread_spec.compressed_sav_file)

        if self.header_block.sram_init_check != 'jk':
            error_msg = (
                "SRAM init check bits incorrect (should be 'jk', was '%s')" %
                (self.header_block.sram_init_check))

            callback(error_msg, current_step, total_steps, False)
            raise ValueError(error_msg)

        self.active_project_number = self.header_block.active_file

        current_step += 1

        callback("Decompressing", current_step, total_steps, True)

        file_blocks = collections.defaultdict(list)

        for block_number, file_number in enumerate(
                self.header_block.block_alloc_table):
            if file_number == self.EMPTY_BLOCK:
                continue

            if file_number < 0 or file_number > 0x1f:
                callback(
                    "File number %x for block %x out of range" %
                    (file_number, block_number),
                    current_step, total_steps, False)

            # The file's header is block 0, so blocks are indexed from 1
            file_blocks[file_number].append(block_number + 1)

        current_step += 1

        for file_number in file_blocks:
            block_numbers = file_blocks[file_number]
            project_size_blks = len(block_numbers)

            block_map = {}

            for block_number in block_numbers:
                offset = self.BLOCKS_START_OFFSET + \
                    (block_number * blockutils.BLOCK_SIZE)

                fp.seek(offset, os.SEEK_SET)

                block_data = bytearray(fp.read(blockutils.BLOCK_SIZE))

                block_map[block_number] = blockutils.Block(block_number,
                                                           block_data)

            reader = BlockReader()
            compressed_data = reader.read(block_map)
            raw_data = filepack.decompress(compressed_data)

            project_name = self.header_block.filenames[file_number]
            project_version = self.header_block.file_versions[file_number]

            callback("Reading project '%s' v%d" %
                     (utils.name_without_zeroes(project_name), project_version),
                     current_step, total_steps, True)

            project = Project(
                name = self.header_block.filenames[file_number],
                version = self.header_block.file_versions[file_number],
                data = raw_data,
                size_blks = project_size_blks)

            self.projects[file_number] = project

            current_step += 1

        for i in xrange(self.NUM_FILES):
            if i not in self.projects:
                self.projects[i] = None

        callback("Import complete!", total_steps, total_steps, True)

    def __str__(self):

        str_stream = StringIO()

        for i, project in self.projects.items():
            if project is not None:
                print >>str_stream, str(project)

        print >>str_stream, "Active Project: %s" % \
            (self.projects[self.active_project_number])

        str_stream_stringval = str_stream.getvalue()
        str_stream.close()
        return str_stream_stringval

    def __eq__(self, other):
        return self.projects == other.projects

    @property
    def project_list(self):
        """The list of :py:class:`pylsdj.Project` s that the .sav file contains"""
        return [(i, self.projects[i]) for i in sorted(self.projects.keys())]

    def _save(self, fp, callback):
        # Marshal 32 possible projects + write preamble + write data + "all done"
        total_steps = 35
        current_step = 0

        writer = BlockWriter()
        factory = BlockFactory()

        # Block allocation table doesn't include header block because it's
        # always in use, so have to add additional block to account for header
        num_blocks = self.BAT_END_OFFSET - self.BAT_START_OFFSET + 2

        header_block = factory.new_block()

        block_table = []

        for i in xrange(num_blocks):
            block_table.append(None)

        # First block is the header block, so we should ignore it when creating
        # the block allocation table
        block_table[0] = -1

        for (i, project) in self.projects.items():
            current_step += 1

            if project is None:
                continue

            callback("Marshaling song '%s'" %
                     (utils.name_without_zeroes(project.name)),
                     current_step - 1, total_steps, True)

            raw_data = project.get_raw_data()
            compressed_data = filepack.compress(raw_data)

            project_block_ids = writer.write(compressed_data, factory)

            for b in project_block_ids:
                block_table[b] = i

        callback("Writing preamble and constructing header block",
                 current_step, total_steps, True)
        current_step += 1
        # Bytes up to START_OFFSET will remain the same
        fp.write(self.preamble)

        # Set header block filenames and versions

        empty_project_name = '\0' * self.FILENAME_LENGTH

        for i, project in self.projects.items():
            if project is None:
                self.header_block.filenames[i] = empty_project_name
                self.header_block.file_versions[i] = 0
            else:
                self.header_block.filenames[i] = project.name
                self.header_block.file_versions[i] = project.version

        self.header_block.active_file = self.active_project_number

        # Ignore the header block when serializing the block allocation table
        for i, b in enumerate(block_table[1:]):
            if b == None:
                file_no = self.EMPTY_BLOCK
            else:
                file_no = b

            self.header_block.block_alloc_table[i] = file_no

        header_block.data = bread.write(
            self.header_block, bread_spec.compressed_sav_file)

        assert len(header_block.data) == blockutils.BLOCK_SIZE, \
            "Header block isn't the expected length; expected 0x%x, got 0x%x" \
            % (blockutils.BLOCK_SIZE, len(header_block.data))

        block_map = factory.blocks

        empty_block_data = []
        for i in xrange(blockutils.BLOCK_SIZE):
            empty_block_data.append(0)

        callback("Writing data to file", current_step, total_steps, True)
        current_step += 1
        for i in xrange(num_blocks):
            if i in block_map:
                data_list = block_map[i].data
            else:
                data_list = empty_block_data

            fp.write(bytearray(data_list))

        callback("Save complete!", total_steps, total_steps, True)

    def save(self, filename, callback=_noop_callback):
        """Save this file.

        :param filename: the file to which to save the .sav file
        :type filename: str
        :param callback: a progress callback function
        :type callback: function
        """
        with open(filename, 'wb') as fp:
            self._save(fp, callback)

if __name__ == "__main__":
    sav = SAVFile(sys.argv[1])
    sav.save(sys.argv[2])



import os
import warnings
import sys
from struct import pack, unpack

def printable_decimal_and_hex(num):
    return "{0:d} (0x{0:x})".format(num)

def add_song_data_property(clazz, property_name, song_data_field_path,
                           use_index=False, doc=None):
    """Add a property to the class `clazz` named `property_name`. This
    property's getter will return the field in self.song.song_data
    corresponding to `song_data_field_path`, and its setter will set that
    field. `song_data_field_path` is a tuple consisting of a sequence of field
    lookups that will lead to the desired field.
    """

    def get_field(self):
        field_val = self.song.song_data

        for field_name in song_data_field_path:
            field_val = getattr(field_val, field_name)

        if use_index:
            return field_val[self.index]
        else:
            return field_val

    def set_field(self, field_val):
        data_obj = self.song.song_data

        for field_name in song_data_field_path[:-1]:
            data_obj = getattr(data_obj, field_name)

        if use_index:
            getattr(data_obj, song_data_field_path[-1])[self.index] = field_val
        else:
            setattr(data_obj, song_data_field_path[-1], field_val)

    setattr(clazz, property_name, property(
        fset=set_field, fget=get_field, doc=doc))

def assert_index_sane(index, upper_bound_exclusive):
    assert type(index) == int, "Indices should be integers; '%s' is not" % (
        index)
    assert 0 <= index < upper_bound_exclusive, (
        "Index %d out of range [%d, %d)" % (index, 0, upper_bound_exclusive))

class ObjectLookupDict(object):
    def __init__(self, id_list, object_list):
        self.id_list = id_list
        self.object_list = object_list

    def __getitem__(self, index):
        assert_index_sane(index, len(self.id_list))

        return self.object_list[self.id_list[index]]

    def __setitem__(self, index, value):
        assert_index_sane(index, len(self.id_list))

        self.id_list[index] = value.index

def name_without_zeroes(name):
    """
    Return a human-readable name without LSDJ's trailing zeroes.
    """
    first_zero = name.find('\0')

    if first_zero == -1:
        return name
    else:
        return str(name[:first_zero])



import json

class SynthSoundParams(object):
    def __init__(self, params):
        self._params = params

    @property
    def volume(self):
        """the wave's volume"""
        return self._params.volume

    @property
    def filter_cutoff(self):
        """the filter's cutoff frequency"""
        return self._params.filter_cutoff

    @property
    def phase_amount(self):
        """the amount of phase shift, ``0`` = no phase,
        ``0x1f`` = maximum phase"""
        return self._params.phase_amount

    @property
    def vertical_shift(self):
        """the amount to shift the waveform vertically"""
        return self._params.vertical_shift


class Synth(object):
    def __init__(self, song, index):
        self._song = song
        self._index = index

        self._params = self._song.song_data.softsynth_params[index]

        self._start = SynthSoundParams(
            self._song.song_data.softsynth_params[index].start)
        self._end = SynthSoundParams(
            self._song.song_data.softsynth_params[index].end)

    @property
    def song(self):
        """the synth's parent Song"""
        return self._song

    @property
    def index(self):
        """the synth's index within its parent song's synth table"""
        return self._index

    @property
    def start(self):
        """parameters for the start of the sound, represented as a
        SynthSoundParams object"""
        return self._start

    @property
    def end(self):
        """parameters for the end of the sound, represented as a
        SynthSoundParams object"""
        return self._end

    @property
    def waveform(self):
        '''the synth\'s waveform type; one of ``"sawtooth"``,
        ``"square"``, ``"sine"``'''
        return self._params.waveform

    @property
    def filter_type(self):
        '''the type of filter applied to the waveform; one of
        ``"lowpass"``, ``"highpass"``, ``"bandpass"``, ``"allpass"``'''
        return self._params.filter_type

    @property
    def filter_resonance(self):
        """boosts the signal around the cutoff
        frequency, to change how bright or dull the wave sounds"""
        return self._params.filter_resonance

    @property
    def distortion(self):
        '''use ``"clip"`` or ``"wrap"`` distortion'''
        return self._params.distortion

    @property
    def phase_type(self):
        '''compresses the waveform horizontally; one of
        ``"normal"``, ``"resync"``, ``"resync2"``'''
        return self._params.phase_type

    @property
    def waves(self):
        """a list of the synth's waveforms, each of which is a list of bytes"""
        return self.song.song_data.wave_frames[self.index]

    def export(self):
        export_struct = {}

        export_struct["params"] = json.loads(self._params.as_json())
        export_struct["waves"] = []

        for wave in self.waves:
            export_struct["waves"].append(list(wave))

        return export_struct

    def import_lsdinst(self, synth_data):
        params_native = self._params.as_native()

        for key in params_native:
            if key[0] == '_':
                continue

            if key in ('start', 'end'):
                self._import_sound_params(
                    synth_data['params'][key], getattr(self, key))
            else:
                setattr(self._params, key, synth_data['params'][key])

        for i, wave in enumerate(synth_data['waves']):
            for j, frame in enumerate(wave):
                self.waves[i][j] = frame

    def _import_sound_params(self, params, dest):
        native_repr = dest.as_native()

        for key in native_repr:
            if key[0] == '_':
                continue

            setattr(dest, key, params[key])



#!/usr/bin/env python

from pylsdj import SAVFile

sav = SAVFile("/Applications/LSDJ/Battery RAM/lsdj.sav")

synth = sav.projects[12].song.synths[2]

print synth.waves
print synth.waveform
print synth.start
print synth._params.waveform



__title__ = 'pylsdj'
__version__ = '1.5.0'
__author__ = 'Alex Rasmussen'
__license__ = 'MIT'
__copyright__ = 'Copyright 2014 Alex Rasmussen'

import bread_spec
import chain
import clock
import consts
import filepack
from instrument import Instrument
from phrase import Phrase
from project import load_lsdsng, load_srm, Project
from savfile import SAVFile
from song import Song, Sequence
from speech_instrument import Word, SpeechInstrument
from synth import Synth, SynthSoundParams
from table import Table, TableFX
import exceptions



from instrument import Instrument
from utils import add_song_data_property

# Max. length of a word name
WORD_NAME_LENGTH = 4

def word_cleanup(word):
    return ''.join([chr(x) for x in word if x != 0]).ljust(WORD_NAME_LENGTH)

class Word(object):
    def __init__(self, song, index):
        self._song = song
        """the word's parent song"""

        self._index = index
        """the word's index within the speech instrument"""

    @property
    def song(self):
        return self._song

    @property
    def index(self):
        return self._index

add_song_data_property(Word, 'name', ("word_names", ), use_index=True,
                       doc="the word's name")
add_song_data_property(Word, 'sounds', ("words", ), use_index=True,
                       doc="a list of the sounds that make up the word; "
                       "each sound has an ``allophone`` and a ``length``")

class SpeechInstrument(object):
    def __init__(self, song):
        self._song = song

        self._words = [
            Word(self._song, i) for i in
            xrange(len(self._song.song_data.words))]

    @property
    def song(self):
        """the speech instrument's parent song"""
        return self._song

    @property
    def words(self):
        """a list of the speech instrument's defined words, as Word objects"""
        return self._words



from utils import assert_index_sane
import json

import bread
import bread_spec

from filepack import DEFAULT_INSTRUMENT
from synth import Synth

from exceptions import ImportException

def new_default_instrument(instr_type):
    instr_data = DEFAULT_INSTRUMENT[:]

    instr_name_to_code = {
        "pulse": 0,
        "wave": 1,
        "kit": 2,
        "noise": 3
    }

    instr_data[0] = instr_name_to_code[instr_type]

    return bread.parse(instr_data, bread_spec.instrument)

class Instrument(object):
    def __init__(self, song, index):
        self.song = song
        self.data = song.song_data.instruments[index]
        self.index = index

    @property
    def name(self):
        """the instrument's name (5 characters, zero-padded)"""
        return self.song.song_data.instrument_names[self.index]

    @name.setter
    def name(self, val):
        self.song.song_data.instrument_names[self.index] = val

    @property
    def type(self):
        """the instrument's type (pulse, wave, kit or noise)"""
        return self.data.instrument_type

    @type.setter
    def type(self, val):
        self.data.instrument_type = val

    def __getattr__(self, name):
        if name == "wave" and self.type == "wave":
            return self.song.song_data.wave_frames[self.index]
        elif name == "table":
            if hasattr(self.data, "table_on") and self.data.table_on:
                assert_index_sane(self.data.table, len(self.song.tables))
                return self.song.tables[self.data.table]
        elif name == "name":
            return self.song.song_data.instrument_names[self.index]
        else:
            return getattr(self.__dict__["data"], name)

    def __setattr__(self, name, value):
        if name == "name":
            self.song.song_data.instrument_names[self.index] = value
        elif name in ["data", "song", "index"]:
            self.__dict__[name] = value
        elif name == "table":
            if not hasattr(self.data, "table_on"):
                raise ValueError("This instrument doesn't have a table")

            self.data.table_on = True
            self.data.table = value.index
        else:
            setattr(self.data, name, value)

    def _get_open_synth_index(self):
        available_synths = set(range(bread_spec.NUM_SYNTHS))

        for instrument in self.song.instruments.as_list():
            if instrument is not None and instrument.type == 'wave':
                available_synths.discard(instrument.synth)

        if len(available_synths) == 0:
            return None
        else:
            return available_synths.pop()

    def import_from_file(self, filename):
        """Import this instrument's settings from the given file. Will
        automatically add the instrument's synth and table to the song's
        synths and tables if needed.

        :param filename: the file from which to load

        :raises ImportException: if importing failed, usually because the song
          doesn't have enough synth or table slots left for the instrument's
          synth or table
        """
        with open(filename, 'r') as fp:
            failure_message = self.import_from_struct(json.load(fp))

            if failure_message is not None:
                raise exceptions.ImportException(failure_message)

        return None

    def import_from_struct(self, lsdinst_struct):
        instr_type = lsdinst_struct['data']['instrument_type']

        # Make sure we've got enough space for the synth if we need one
        if instr_type == 'wave':
            synth_index = self._get_open_synth_index()

            if synth_index is None:
                return ("No available synth slot in which to store the "
                        "instrument's synth data")

        # Make sure we've got enough room for the table if we need it
        if 'table' in lsdinst_struct:
            table_index = self.song.tables.next_free()

            if table_index is None:
                return ("No available table slot in which to store the "
                        "instrument's table data")
            else:
                self.song.tables.allocate(table_index)

        self.name = lsdinst_struct['name']

        # If this instrument is of a different type than we're currently
        # storing, we've got to make a new one of the appropriate type into
        # which to demarshal
        if self.type != instr_type:
            self.data = new_default_instrument(instr_type)
            self.song.song_data.instruments[self.index] = self.data

        native_repr = self.data.as_native()
        self._import_instr_data(lsdinst_struct['data'], native_repr, self.data)

        if instr_type == 'wave':
            self.data.synth = synth_index

            synth = Synth(self.song, synth_index)
            synth.import_lsdinst(lsdinst_struct['synth'])

        if 'table' in lsdinst_struct:
            table = self.song.tables[table_index]

            self.data.table_on = True
            self.data.table = table_index
            table.import_lsdinst(lsdinst_struct['table'])


    def _import_instr_data(self, import_data, native_repr, output_data):
        for (key, val) in native_repr.items():
            if key[0] == '_' or key in ('synth'):
                continue

            if type(val) == dict:
                self._import_instr_data(
                    import_data[key], val, getattr(output_data, key))
            else:
                setattr(output_data, key, import_data[key])

    def export_to_file(self, filename):
        """Export this instrument's settings to a file.

        :param filename: the name of the file
        """
        instr_json = self.export_struct()

        with open(filename, 'w') as fp:
            json.dump(instr_json, fp, indent=2)

    def export_struct(self):
        export_struct = {}

        export_struct['name'] = self.name
        export_struct['data'] = {}

        data_json = json.loads(self.data.as_json())

        for key, value in data_json.items():
            if key[0] != '_' or key in ('synth','table'):
                export_struct['data'][key] = value

        if self.type == 'wave':
            export_struct['synth'] = Synth(self.song, self.synth).export()

        if self.table is not None:
            export_struct['table'] = self.table.export()
        return export_struct
